#!/usr/bin/env python
# coding: utf-8

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import struct
from unittest import TestCase

from neo4j.data import DataDehydrator
from neo4j.packstream import Packer
from neo4j.spatial import (
    Point,
    point_type,
)


class PointTestCase(TestCase):

    def test_wrong_type_arguments(self):
        for argument in (("a", "b"), ({"x": 1.0, "y": 2.0})):
            with self.subTest():
                with self.assertRaises(ValueError):
                    Point(argument)

    def test_number_arguments(self):
        for argument in ((1, 2), (1.2, 2.1)):
            with self.subTest():
                p = Point(argument)
                assert tuple(p) == argument

    def test_dehydration(self):
        MyPoint = point_type("MyPoint", ["x", "y"], {2: 1234})
        coordinates = (.1, 0)
        p = MyPoint(coordinates)

        dehydrator = DataDehydrator()
        buffer = io.BytesIO()
        packer = Packer(buffer)
        packer.pack(dehydrator.dehydrate((p,))[0])
        self.assertEqual(
            buffer.getvalue(),
            b"\xB3X" +
            b"\xC9" + struct.pack(">h", 1234) +
            b"".join(map(lambda c: b"\xC1" + struct.pack(">d", c), coordinates))
        )

    def test_immutable_coordinates(self):
        MyPoint = point_type("MyPoint", ["x", "y"], {2: 1234})
        coordinates = (.1, 0)
        p = MyPoint(coordinates)
        with self.assertRaises(AttributeError):
            p.x = 2.0
        with self.assertRaises(AttributeError):
            p.y = 2.0
        with self.assertRaises(TypeError):
            p[0] = 2.0
        with self.assertRaises(TypeError):
            p[1] = 2.0



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

# python -m pytest tests/unit/test_import_neo4j.py -s -v


def test_import_dunder_version():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_dunder_version
    from neo4j import __version__


def test_import_graphdatabase():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_graphdatabase
    from neo4j import GraphDatabase


def test_import_driver():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_driver
    from neo4j import Driver


def test_import_boltdriver():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_boltdriver
    from neo4j import BoltDriver


def test_import_neo4jdriver():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_neo4jdriver
    from neo4j import Neo4jDriver


def test_import_auth():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_auth
    from neo4j import Auth


def test_import_authtoken():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_authtoken
    from neo4j import AuthToken


def test_import_basic_auth():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_auth
    from neo4j import basic_auth


def test_import_kerberos_auth():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_kerberos_auth
    from neo4j import kerberos_auth


def test_import_custom_auth():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_custom_auth
    from neo4j import custom_auth


def test_import_read_access():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_read_access
    from neo4j import READ_ACCESS


def test_import_write_access():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_write_access
    from neo4j import WRITE_ACCESS


def test_import_transaction():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_transaction
    from neo4j import Transaction


def test_import_record():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_record
    from neo4j import Record


def test_import_session():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_session
    from neo4j import Session


def test_import_sessionconfig():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_sessionconfig
    from neo4j import SessionConfig


def test_import_query():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_query
    from neo4j import Query


def test_import_result():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_result
    from neo4j import Result


def test_import_resultsummary():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_resultsummary
    from neo4j import ResultSummary


def test_import_unit_of_work():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_unit_of_work
    from neo4j import unit_of_work


def test_import_config():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_config
    from neo4j import Config


def test_import_poolconfig():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_poolconfig
    from neo4j import PoolConfig


def test_import_graph():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_graph
    import neo4j.graph as graph


def test_import_graph_node():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_graph_node
    from neo4j.graph import Node


def test_import_graph_path():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_graph_path
    from neo4j.graph import Path


def test_import_graph_graph():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_graph_graph
    from neo4j.graph import Graph


def test_import_spatial():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_spatial
    import neo4j.spatial as spatial


def test_import_time():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_time
    import neo4j.time as time


def test_import_exceptions():
    # python -m pytest tests/unit/test_import_neo4j.py -s -v -k test_import_exceptions
    import neo4j.exceptions as exceptions





#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from uuid import uuid4
from unittest.mock import (
    MagicMock,
    NonCallableMagicMock,
)

import pytest

from neo4j import (
    Query,
    Transaction,
)

from ._fake_connection import fake_connection


@pytest.mark.parametrize(("explicit_commit", "close"), (
    (False, False),
    (True, False),
    (True, True),
))
def test_transaction_context_when_committing(mocker, fake_connection,
                                             explicit_commit, close):
    on_closed = MagicMock()
    on_error = MagicMock()
    tx = Transaction(fake_connection, 2, on_closed, on_error)
    mock_commit = mocker.patch.object(tx, "commit", wraps=tx.commit)
    mock_rollback = mocker.patch.object(tx, "rollback", wraps=tx.rollback)
    with tx as tx_:
        assert mock_commit.call_count == 0
        assert mock_rollback.call_count == 0
        assert tx is tx_
        if explicit_commit:
            tx_.commit()
            mock_commit.assert_called_once_with()
            assert tx.closed()
        if close:
            tx_.close()
            assert tx_.closed()
    mock_commit.assert_called_once_with()
    assert mock_rollback.call_count == 0
    assert tx_.closed()


@pytest.mark.parametrize(("rollback", "close"), (
    (True, False),
    (False, True),
    (True, True),
))
def test_transaction_context_with_explicit_rollback(mocker, fake_connection,
                                                    rollback, close):
    on_closed = MagicMock()
    on_error = MagicMock()
    tx = Transaction(fake_connection, 2, on_closed, on_error)
    mock_commit = mocker.patch.object(tx, "commit", wraps=tx.commit)
    mock_rollback = mocker.patch.object(tx, "rollback", wraps=tx.rollback)
    with tx as tx_:
        assert mock_commit.call_count == 0
        assert mock_rollback.call_count == 0
        assert tx is tx_
        if rollback:
            tx_.rollback()
            mock_rollback.assert_called_once_with()
            assert tx_.closed()
        if close:
            tx_.close()
            mock_rollback.assert_called_once_with()
            assert tx_.closed()
    assert mock_commit.call_count == 0
    mock_rollback.assert_called_once_with()
    assert tx_.closed()


def test_transaction_context_calls_rollback_on_error(mocker, fake_connection):
    class OopsError(RuntimeError):
        pass

    on_closed = MagicMock()
    on_error = MagicMock()
    tx = Transaction(fake_connection, 2, on_closed, on_error)
    mock_commit = mocker.patch.object(tx, "commit", wraps=tx.commit)
    mock_rollback = mocker.patch.object(tx, "rollback", wraps=tx.rollback)
    with pytest.raises(OopsError):
        with tx as tx_:
            assert mock_commit.call_count == 0
            assert mock_rollback.call_count == 0
            assert tx is tx_
            raise OopsError
    assert mock_commit.call_count == 0
    mock_rollback.assert_called_once_with()
    assert tx_.closed()


@pytest.mark.parametrize(("parameters", "error_type"), (
    # maps must have string keys
    ({"x": {1: 'eins', 2: 'zwei', 3: 'drei'}}, TypeError),
    ({"x": {(1, 2): '1+2i', (2, 0): '2'}}, TypeError),
    ({"x": uuid4()}, TypeError),
))
def test_transaction_run_with_invalid_parameters(fake_connection, parameters,
                                                 error_type):
    on_closed = MagicMock()
    on_error = MagicMock()
    tx = Transaction(fake_connection, 2, on_closed, on_error)
    with pytest.raises(error_type):
        tx.run("RETURN $x", **parameters)


def test_transaction_run_takes_no_query_object(fake_connection):
    on_closed = MagicMock()
    on_error = MagicMock()
    tx = Transaction(fake_connection, 2, on_closed, on_error)
    with pytest.raises(ValueError):
        tx.run(Query("RETURN 1"))


def test_transaction_rollbacks_on_open_connections(fake_connection):
    tx = Transaction(fake_connection, 2,
                     lambda *args, **kwargs: None,
                     lambda *args, **kwargs: None)
    with tx as tx_:
        fake_connection.is_reset_mock.return_value = False
        fake_connection.is_reset_mock.reset_mock()
        tx_.rollback()
        fake_connection.is_reset_mock.assert_called_once()
        fake_connection.reset.assert_not_called()
        fake_connection.rollback.assert_called_once()


def test_transaction_no_rollback_on_reset_connections(fake_connection):
    tx = Transaction(fake_connection, 2,
                     lambda *args, **kwargs: None,
                     lambda *args, **kwargs: None)
    with tx as tx_:
        fake_connection.is_reset_mock.return_value = True
        fake_connection.is_reset_mock.reset_mock()
        tx_.rollback()
        fake_connection.is_reset_mock.assert_called_once()
        fake_connection.reset.asset_not_called()
        fake_connection.rollback.asset_not_called()


def test_transaction_no_rollback_on_closed_connections(fake_connection):
    tx = Transaction(fake_connection, 2,
                     lambda *args, **kwargs: None,
                     lambda *args, **kwargs: None)
    with tx as tx_:
        fake_connection.closed.return_value = True
        fake_connection.closed.reset_mock()
        fake_connection.is_reset_mock.reset_mock()
        tx_.rollback()
        fake_connection.closed.assert_called_once()
        fake_connection.is_reset_mock.asset_not_called()
        fake_connection.reset.asset_not_called()
        fake_connection.rollback.asset_not_called()


def test_transaction_no_rollback_on_defunct_connections(fake_connection):
    tx = Transaction(fake_connection, 2,
                     lambda *args, **kwargs: None,
                     lambda *args, **kwargs: None)
    with tx as tx_:
        fake_connection.defunct.return_value = True
        fake_connection.defunct.reset_mock()
        fake_connection.is_reset_mock.reset_mock()
        tx_.rollback()
        fake_connection.defunct.assert_called_once()
        fake_connection.is_reset_mock.asset_not_called()
        fake_connection.reset.asset_not_called()
        fake_connection.rollback.asset_not_called()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from contextlib import redirect_stdout
from io import StringIO

# tag::cypher-error-import[]
from neo4j.exceptions import ClientError
# end::cypher-error-import[]


class Neo4jErrorExample:

    def __init__(self, driver):
        self.driver = driver

    # tag::cypher-error[]
    def get_employee_number(self, name):
        with self.driver.session() as session:
            try:
                session.read_transaction(self.select_employee, name)
            except ClientError as error:
                print(error.message)
                return -1

    @staticmethod
    def select_employee(tx, name):
        result = tx.run("SELECT * FROM Employees WHERE name = $name", name=name)
        return result.single()["employee_number"]
    # end::cypher-error[]


def test_example(bolt_driver):
    s = StringIO()
    with redirect_stdout(s):
        example = Neo4jErrorExample(bolt_driver)
        example.get_employee_number('Alice')
        assert s.getvalue().startswith("Invalid input")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
import datetime

from pytz import (
    FixedOffset,
    timezone,
    utc,
)

from neo4j.exceptions import CypherTypeError
from neo4j.time import (
    Date,
    Time,
    DateTime,
    Duration,
)


def test_native_date_input(cypher_eval):
    from datetime import date
    result = cypher_eval("CYPHER runtime=interpreted WITH $x AS x "
                         "RETURN [x.year, x.month, x.day]",
                         x=date(1976, 6, 13))
    year, month, day = result
    assert year == 1976
    assert month == 6
    assert day == 13


def test_date_input(cypher_eval):
    result = cypher_eval("CYPHER runtime=interpreted WITH $x AS x "
                         "RETURN [x.year, x.month, x.day]",
                         x=Date(1976, 6, 13))
    year, month, day = result
    assert year == 1976
    assert month == 6
    assert day == 13


def test_date_array_input(cypher_eval):
    data = [DateTime.now().date(), Date(1976, 6, 13)]
    value = cypher_eval("CREATE (a {x:$x}) RETURN a.x", x=data)
    assert value == data


def test_native_time_input(cypher_eval):
    from datetime import time
    result = cypher_eval("CYPHER runtime=interpreted WITH $x AS x "
                         "RETURN [x.hour, x.minute, x.second, x.nanosecond]",
                         x=time(12, 34, 56, 789012))
    hour, minute, second, nanosecond = result
    assert hour == 12
    assert minute == 34
    assert second == 56
    assert nanosecond == 789012000


def test_whole_second_time_input(cypher_eval):
    result = cypher_eval("CYPHER runtime=interpreted WITH $x AS x "
                         "RETURN [x.hour, x.minute, x.second]",
                         x=Time(12, 34, 56))
    hour, minute, second = result
    assert hour == 12
    assert minute == 34
    assert second == 56


def test_nanosecond_resolution_time_input(cypher_eval):
    result = cypher_eval("CYPHER runtime=interpreted WITH $x AS x "
                         "RETURN [x.hour, x.minute, x.second, x.nanosecond]",
                         x=Time(12, 34, 56.789012345))
    hour, minute, second, nanosecond = result
    assert hour == 12
    assert minute == 34
    assert second == 56
    assert nanosecond == 789012345


def test_time_with_numeric_time_offset_input(cypher_eval):
    result = cypher_eval("CYPHER runtime=interpreted WITH $x AS x "
                         "RETURN [x.hour, x.minute, x.second, "
                         "        x.nanosecond, x.offset]",
                         x=Time(12, 34, 56.789012345, tzinfo=FixedOffset(90)))
    hour, minute, second, nanosecond, offset = result
    assert hour == 12
    assert minute == 34
    assert second == 56
    assert nanosecond == 789012345
    assert offset == "+01:30"


def test_time_array_input(cypher_eval):
    data = [Time(12, 34, 56), Time(10, 0, 0)]
    value = cypher_eval("CREATE (a {x:$x}) RETURN a.x", x=data)
    assert value == data


def test_native_datetime_input(cypher_eval):
    from datetime import datetime
    result = cypher_eval("CYPHER runtime=interpreted WITH $x AS x "
                         "RETURN [x.year, x.month, x.day, "
                         "        x.hour, x.minute, x.second, x.nanosecond]",
                         x=datetime(1976, 6, 13, 12, 34, 56, 789012))
    year, month, day, hour, minute, second, nanosecond = result
    assert year == 1976
    assert month == 6
    assert day == 13
    assert hour == 12
    assert minute == 34
    assert second == 56
    assert nanosecond == 789012000


def test_whole_second_datetime_input(cypher_eval):
    result = cypher_eval("CYPHER runtime=interpreted WITH $x AS x "
                         "RETURN [x.year, x.month, x.day, "
                         "        x.hour, x.minute, x.second]",
                         x=DateTime(1976, 6, 13, 12, 34, 56))
    year, month, day, hour, minute, second = result
    assert year == 1976
    assert month == 6
    assert day == 13
    assert hour == 12
    assert minute == 34
    assert second == 56


def test_nanosecond_resolution_datetime_input(cypher_eval):
    result = cypher_eval("CYPHER runtime=interpreted WITH $x AS x "
                         "RETURN [x.year, x.month, x.day, "
                         "        x.hour, x.minute, x.second, x.nanosecond]",
                         x=DateTime(1976, 6, 13, 12, 34, 56.789012345))
    year, month, day, hour, minute, second, nanosecond = result
    assert year == 1976
    assert month == 6
    assert day == 13
    assert hour == 12
    assert minute == 34
    assert second == 56
    assert nanosecond == 789012345


def test_datetime_with_numeric_time_offset_input(cypher_eval):
    result = cypher_eval("CYPHER runtime=interpreted WITH $x AS x "
                         "RETURN [x.year, x.month, x.day, "
                         "        x.hour, x.minute, x.second, "
                         "        x.nanosecond, x.offset]",
                         x=DateTime(1976, 6, 13, 12, 34, 56.789012345,
                                    tzinfo=FixedOffset(90)))
    year, month, day, hour, minute, second, nanosecond, offset = result
    assert year == 1976
    assert month == 6
    assert day == 13
    assert hour == 12
    assert minute == 34
    assert second == 56
    assert nanosecond == 789012345
    assert offset == "+01:30"


def test_datetime_with_named_time_zone_input(cypher_eval):
    dt = DateTime(1976, 6, 13, 12, 34, 56.789012345)
    input_value = timezone("US/Pacific").localize(dt)
    result = cypher_eval("CYPHER runtime=interpreted WITH $x AS x "
                         "RETURN [x.year, x.month, x.day, "
                         "        x.hour, x.minute, x.second, "
                         "        x.nanosecond, x.timezone]",
                         x=input_value)
    year, month, day, hour, minute, second, nanosecond, tz = result
    assert year == input_value.year
    assert month == input_value.month
    assert day == input_value.day
    assert hour == input_value.hour
    assert minute == input_value.minute
    assert second == int(input_value.second)
    assert nanosecond == int(1000000000 * input_value.second % 1000000000)
    assert tz == input_value.tzinfo.zone


def test_datetime_array_input(cypher_eval):
    data = [DateTime(2018, 4, 6, 13, 4, 42.516120), DateTime(1976, 6, 13)]
    value = cypher_eval("CREATE (a {x:$x}) RETURN a.x", x=data)
    assert value == data


def test_duration_input(cypher_eval):
    result = cypher_eval("CYPHER runtime=interpreted WITH $x AS x "
                         "RETURN [x.months, x.days, x.seconds, "
                         "        x.microsecondsOfSecond]",
                         x=Duration(years=1, months=2, days=3, hours=4,
                                    minutes=5, seconds=6.789012))
    months, days, seconds, microseconds = result
    assert months == 14
    assert days == 3
    assert seconds == 14706
    assert microseconds == 789012


def test_duration_array_input(cypher_eval):
    data = [Duration(1, 2, 3, 4, 5, 6), Duration(9, 8, 7, 6, 5, 4)]
    value = cypher_eval("CREATE (a {x:$x}) RETURN a.x", x=data)
    assert value == data


def test_timedelta_input(cypher_eval):
    from datetime import timedelta
    result = cypher_eval("CYPHER runtime=interpreted WITH $x AS x "
                         "RETURN [x.months, x.days, x.seconds, "
                         "        x.microsecondsOfSecond]",
                         x=timedelta(days=3, hours=4, minutes=5,
                                     seconds=6.789012))
    months, days, seconds, microseconds = result
    assert months == 0
    assert days == 3
    assert seconds == 14706
    assert microseconds == 789012


def test_mixed_array_input(cypher_eval):
    data = [Date(1976, 6, 13), Duration(9, 8, 7, 6, 5, 4)]
    with pytest.raises(CypherTypeError):
        _ = cypher_eval("CREATE (a {x:$x}) RETURN a.x", x=data)


def test_date_output(cypher_eval):
    value = cypher_eval("RETURN date('1976-06-13')")
    assert isinstance(value, Date)
    assert value == Date(1976, 6, 13)


def test_whole_second_time_output(cypher_eval):
    value = cypher_eval("RETURN time('12:34:56')")
    assert isinstance(value, Time)
    assert value == Time(12, 34, 56, tzinfo=FixedOffset(0))


def test_nanosecond_resolution_time_output(cypher_eval):
    value = cypher_eval("RETURN time('12:34:56.789012345')")
    assert isinstance(value, Time)
    assert value == Time(12, 34, 56.789012345, tzinfo=FixedOffset(0))


def test_time_with_numeric_time_offset_output(cypher_eval):
    value = cypher_eval("RETURN time('12:34:56.789012345+0130')")
    assert isinstance(value, Time)
    assert value == Time(12, 34, 56.789012345, tzinfo=FixedOffset(90))


def test_whole_second_localtime_output(cypher_eval):
    value = cypher_eval("RETURN localtime('12:34:56')")
    assert isinstance(value, Time)
    assert value == Time(12, 34, 56)


def test_nanosecond_resolution_localtime_output(cypher_eval):
    value = cypher_eval("RETURN localtime('12:34:56.789012345')")
    assert isinstance(value, Time)
    assert value == Time(12, 34, 56.789012345)


def test_whole_second_datetime_output(cypher_eval):
    value = cypher_eval("RETURN datetime('1976-06-13T12:34:56')")
    assert isinstance(value, DateTime)
    assert value == DateTime(1976, 6, 13, 12, 34, 56, tzinfo=utc)


def test_nanosecond_resolution_datetime_output(cypher_eval):
    value = cypher_eval("RETURN datetime('1976-06-13T12:34:56.789012345')")
    assert isinstance(value, DateTime)
    assert value == DateTime(1976, 6, 13, 12, 34, 56.789012345, tzinfo=utc)


def test_datetime_with_numeric_time_offset_output(cypher_eval):
    value = cypher_eval("RETURN "
                        "datetime('1976-06-13T12:34:56.789012345+01:30')")
    assert isinstance(value, DateTime)
    assert value == DateTime(1976, 6, 13, 12, 34, 56.789012345,
                             tzinfo=FixedOffset(90))


def test_datetime_with_named_time_zone_output(cypher_eval):
    value = cypher_eval("RETURN datetime('1976-06-13T12:34:56.789012345"
                        "[Europe/London]')")
    assert isinstance(value, DateTime)
    dt = DateTime(1976, 6, 13, 12, 34, 56.789012345)
    assert value == timezone("Europe/London").localize(dt)


def test_whole_second_localdatetime_output(cypher_eval):
    value = cypher_eval("RETURN localdatetime('1976-06-13T12:34:56')")
    assert isinstance(value, DateTime)
    assert value == DateTime(1976, 6, 13, 12, 34, 56)


def test_nanosecond_resolution_localdatetime_output(cypher_eval):
    value = cypher_eval("RETURN "
                        "localdatetime('1976-06-13T12:34:56.789012345')")
    assert isinstance(value, DateTime)
    assert value == DateTime(1976, 6, 13, 12, 34, 56.789012345)


def test_duration_output(cypher_eval):
    value = cypher_eval("RETURN duration('P1Y2M3DT4H5M6.789S')")
    assert isinstance(value, Duration)
    assert value == Duration(years=1, months=2, days=3, hours=4,
                             minutes=5, seconds=6.789)


def test_nanosecond_resolution_duration_output(cypher_eval):
    value = cypher_eval("RETURN duration('P1Y2M3DT4H5M6.789123456S')")
    assert isinstance(value, Duration)
    assert value == Duration(years=1, months=2, days=3, hours=4,
                             minutes=5, seconds=6.789123456)


def test_datetime_parameter_case1(session):
    # python -m pytest tests/integration/test_temporal_types.py -s -v -k test_datetime_parameter_case1
    dt1 = session.run("RETURN datetime('2019-10-30T07:54:02.129790001+00:00')").single().value()
    assert isinstance(dt1, DateTime)

    dt2 = session.run("RETURN $date_time", date_time=dt1).single().value()
    assert isinstance(dt2, DateTime)

    assert dt1 == dt2


def test_datetime_parameter_case2(session):
    # python -m pytest tests/integration/test_temporal_types.py -s -v -k test_datetime_parameter_case2
    dt1 = session.run("RETURN datetime('2019-10-30T07:54:02.129790999[UTC]')").single().value()
    assert isinstance(dt1, DateTime)
    assert dt1.iso_format() == "2019-10-30T07:54:02.129790999+00:00"

    dt2 = session.run("RETURN $date_time", date_time=dt1).single().value()
    assert isinstance(dt2, DateTime)

    assert dt1 == dt2


def test_datetime_parameter_case3(session):
    # python -m pytest tests/integration/test_temporal_types.py -s -v -k test_datetime_parameter_case1
    dt1 = session.run("RETURN datetime('2019-10-30T07:54:02.129790+00:00')").single().value()
    assert isinstance(dt1, DateTime)

    dt2 = session.run("RETURN $date_time", date_time=dt1).single().value()
    assert isinstance(dt2, DateTime)

    assert dt1 == dt2


def test_time_parameter_case1(session):
    # python -m pytest tests/integration/test_temporal_types.py -s -v -k test_time_parameter_case1
    t1 = session.run("RETURN time('07:54:02.129790001+00:00')").single().value()
    assert isinstance(t1, Time)

    t2 = session.run("RETURN $time", time=t1).single().value()
    assert isinstance(t2, Time)

    assert t1 == t2


def test_time_parameter_case2(session):
    # python -m pytest tests/integration/test_temporal_types.py -s -v -k test_time_parameter_case2
    t1 = session.run("RETURN time('07:54:02.129790999+00:00')").single().value()
    assert isinstance(t1, Time)
    assert t1.iso_format() == "07:54:02.129790999+00:00"
    time_zone_delta = t1.utc_offset()
    assert isinstance(time_zone_delta, datetime.timedelta)
    assert time_zone_delta == datetime.timedelta(0)

    t2 = session.run("RETURN $time", time=t1).single().value()
    assert isinstance(t2, Time)

    assert t1 == t2


def test_time_parameter_case3(session):
    # python -m pytest tests/integration/test_temporal_types.py -s -v -k test_time_parameter_case3
    t1 = session.run("RETURN time('07:54:02.129790+00:00')").single().value()
    assert isinstance(t1, Time)

    t2 = session.run("RETURN $time", time=t1).single().value()
    assert isinstance(t2, Time)

    assert t1 == t2


#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

from neo4j import (
    GraphDatabase,
    Neo4jDriver,
    TRUST_ALL_CERTIFICATES,
    TRUST_SYSTEM_CA_SIGNED_CERTIFICATES,
    DEFAULT_DATABASE,
)
from neo4j.api import (
    READ_ACCESS,
    WRITE_ACCESS,
)
from neo4j.exceptions import (
    ServiceUnavailable,
    TransientError,
    SessionExpired,
    ConfigurationError,
)
from neo4j._exceptions import (
    BoltSecurityError,
)
from tests.stub.conftest import StubCluster

# python -m pytest tests/stub/test_routingdriver.py -s -v
# TODO: those tests will stay until a uniform behavior across the drivers has
#       been specified and tests are created in testkit
@pytest.mark.parametrize(
    "test_script",
    [
        "v3/router.script",
        "v4x0/router_port_9001_one_read_port_9004_one_write_port_9006.script",
    ]
)
def test_neo4j_driver_verify_connectivity(driver_info, test_script):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_neo4j_driver_verify_connectivity
    with StubCluster(test_script):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"], user_agent="test") as driver:
            assert isinstance(driver, Neo4jDriver)
            assert driver.verify_connectivity() is not None


# @pytest.mark.skip(reason="Flaky")
@pytest.mark.parametrize(
    "test_script",
    [
        "v3/router.script",
        "v4x0/router_port_9001_one_read_port_9004_one_write_port_9006.script",
    ]
)
def test_neo4j_driver_verify_connectivity_server_down(driver_info, test_script):
    # python -m pytest tests/stub/test_routingdriver.py -s -v -k test_neo4j_driver_verify_connectivity_server_down
    with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"], user_agent="test") as driver:
        assert isinstance(driver, Neo4jDriver)

        with pytest.raises(ServiceUnavailable):
            driver.verify_connectivity()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from unittest import TestCase, SkipTest

try:
    from urllib.request import urlretrieve
except ImportError:
    from urllib import urlretrieve

from neo4j import GraphDatabase
from neo4j.exceptions import AuthError

from tests.env import NEO4J_USER, NEO4J_PASSWORD, NEO4J_SERVER_URI


def is_listening(address):
    from socket import create_connection
    try:
        s = create_connection(address)
    except IOError:
        return False
    else:
        s.close()
        return True


class RemoteGraphDatabaseServer(object):
    server_uri = NEO4J_SERVER_URI or "bolt://localhost:7687"
    auth_token = (NEO4J_USER or "neo4j", NEO4J_PASSWORD)
    encrypted = NEO4J_SERVER_URI is not None

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()

    @classmethod
    def start(cls):
        with GraphDatabase.driver(cls.server_uri, auth=cls.auth_token, encrypted=cls.encrypted) as driver:
            try:
                with driver.session():
                    print("Using existing remote server {}\n".format(cls.server_uri))
                    return
            except AuthError as error:
                raise RuntimeError("Failed to authenticate (%s)" % error)
        raise SkipTest("No remote Neo4j server available for %s" % cls.__name__)

    @classmethod
    def stop(cls):
        pass



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

from neo4j.exceptions import (
    ServiceUnavailable,
    ConfigurationError,
    UnsupportedServerProduct,
)
from neo4j._exceptions import (
    BoltHandshakeError,
    BoltSecurityError,
)

from neo4j import (
    GraphDatabase,
    BoltDriver,
    Query,
    WRITE_ACCESS,
    READ_ACCESS,
    TRUST_ALL_CERTIFICATES,
    TRUST_SYSTEM_CA_SIGNED_CERTIFICATES,
    DEFAULT_DATABASE,
    Result,
    unit_of_work,
    Transaction,
)

from tests.stub.conftest import (
    StubCluster,
)

# python -m pytest tests/stub/test_directdriver.py -s -v


driver_config = {
    "encrypted": False,
    "user_agent": "test",
    "max_connection_lifetime": 1000,
    "max_connection_pool_size": 10,
    "keep_alive": True,
    "resolver": None,
}


session_config = {
    "default_access_mode": READ_ACCESS,
    "connection_acquisition_timeout": 1.0,
    "max_transaction_retry_time": 1.0,
    "initial_retry_delay": 1.0,
    "retry_delay_multiplier": 1.0,
    "retry_delay_jitter_factor": 0.1,
    "fetch_size": -1,
}


# TODO: those tests will stay until a uniform behavior across the drivers has
#       been specified and tests are created in testkit
def test_direct_driver_with_wrong_port(driver_info):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_direct_driver_with_wrong_port
    uri = "bolt://127.0.0.1:9002"
    with pytest.raises(ServiceUnavailable):
        driver = GraphDatabase.driver(uri, auth=driver_info["auth_token"], **driver_config)
        driver.verify_connectivity()


@pytest.mark.parametrize(
    "test_script, test_expected",
    [
        ("v3/return_1_port_9001.script", "Neo4j/3.0.0"),
        ("v4x0/return_1_port_9001.script", "Neo4j/4.0.0"),
    ]
)
def test_direct_verify_connectivity(driver_info, test_script, test_expected):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_direct_verify_connectivity
    with StubCluster(test_script):
        uri = "bolt://localhost:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"], **driver_config) as driver:
            assert isinstance(driver, BoltDriver)
            assert driver.verify_connectivity(default_access_mode=READ_ACCESS) == test_expected


@pytest.mark.parametrize(
    "test_script",
    [
        "v3/disconnect_on_run.script",
        "v4x0/disconnect_on_run.script",
    ]
)
def test_direct_verify_connectivity_disconnect_on_run(driver_info, test_script):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_direct_verify_connectivity_disconnect_on_run
    with StubCluster(test_script):
        uri = "bolt://127.0.0.1:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"], **driver_config) as driver:
            with pytest.raises(ServiceUnavailable):
                driver.verify_connectivity(default_access_mode=READ_ACCESS)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from neo4j.conf import WorkspaceConfig
from neo4j.exceptions import ServiceUnavailable
from neo4j.io import Neo4jPool


class Workspace:

    def __init__(self, pool, config):
        assert isinstance(config, WorkspaceConfig)
        self._pool = pool
        self._config = config
        self._connection = None
        self._connection_access_mode = None
        # Sessions are supposed to cache the database on which to operate.
        self._cached_database = False
        self._bookmarks = None

    def __del__(self):
        try:
            self.close()
        except OSError:
            pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def _set_cached_database(self, database):
        self._cached_database = True
        self._config.database = database

    def _connect(self, access_mode):
        if self._connection:
            # TODO: Investigate this
            # log.warning("FIXME: should always disconnect before connect")
            self._connection.send_all()
            self._connection.fetch_all()
            self._disconnect()
        if not self._cached_database:
            if (self._config.database is not None
                    or not isinstance(self._pool, Neo4jPool)):
                self._set_cached_database(self._config.database)
            else:
                # This is the first time we open a connection to a server in a
                # cluster environment for this session without explicitly
                # configured database. Hence, we request a routing table update
                # to try to fetch the home database. If provided by the server,
                # we shall use this database explicitly for all subsequent
                # actions within this session.
                self._pool.update_routing_table(
                    database=self._config.database,
                    imp_user=self._config.impersonated_user,
                    bookmarks=self._bookmarks,
                    database_callback=self._set_cached_database
                )
        self._connection = self._pool.acquire(
            access_mode=access_mode,
            timeout=self._config.connection_acquisition_timeout,
            database=self._config.database,
            bookmarks=self._bookmarks
        )
        self._connection_access_mode = access_mode

    def _disconnect(self, sync=False):
        if self._connection:
            if sync:
                try:
                    self._connection.send_all()
                    self._connection.fetch_all()
                except (WorkspaceError, ServiceUnavailable):
                    pass
            if self._connection:
                self._pool.release(self._connection)
                self._connection = None
            self._connection_access_mode = None

    def close(self):
        self._disconnect(sync=True)


class WorkspaceError(Exception):

    pass



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import time
from decimal import Decimal

import pytest
from pytz import (
    build_tzinfo,
    timezone,
    FixedOffset,
)

from neo4j.time import Time
from neo4j.time.arithmetic import (
    nano_add,
    nano_div,
)


timezone_us_eastern = timezone("US/Eastern")
timezone_utc = timezone("UTC")


def seconds_options(seconds, nanoseconds):
    yield seconds, nanoseconds
    yield seconds + nanoseconds / 1000000000,


class TestTime:

    def test_bad_attribute(self):
        t = Time(12, 34, 56.789)
        with pytest.raises(AttributeError):
            _ = t.x

    def test_simple_time(self):
        t = Time(12, 34, 56.789)
        assert t.hour_minute_second == (12, 34, Decimal("56.789"))
        assert t.hour_minute_second_nanosecond == (12, 34, 56, 789000000)
        assert t.ticks == 45296.789
        assert t.hour == 12
        assert t.minute == 34
        assert t.second == Decimal("56.789")

    def test_midnight(self):
        t = Time(0, 0, 0)
        assert t.hour_minute_second == (0, 0, 0)
        assert t.ticks == 0
        assert t.hour == 0
        assert t.minute == 0
        assert t.second == 0

    def test_nanosecond_precision(self):
        t = Time(12, 34, 56.789123456)
        assert t.hour_minute_second == (12, 34, Decimal("56.789123456"))
        assert t.ticks == 45296.789123456
        assert t.hour == 12
        assert t.minute == 34
        assert t.second == Decimal("56.789123456")

    def test_str(self):
        t = Time(12, 34, 56, 789123456)
        assert str(t) == "12:34:56.789123456"

    def test_now_without_tz(self):
        t = Time.now()
        assert isinstance(t, Time)

    def test_now_with_tz(self):
        t = Time.now(tz=timezone_us_eastern)
        assert isinstance(t, Time)
        assert t.tzinfo == timezone_us_eastern

    def test_utc_now(self):
        t = Time.utc_now()
        assert isinstance(t, Time)

    def test_from_native(self):
        native = time(12, 34, 56, 789123)
        t = Time.from_native(native)
        assert t.hour == native.hour
        assert t.minute == native.minute
        assert t.second == \
               Decimal(native.second) + Decimal(native.microsecond) / 1000000

    def test_to_native(self):
        t = Time(12, 34, 56.789123456)
        native = t.to_native()
        assert t.hour == native.hour
        assert t.minute == native.minute
        assert 56.789123 == nano_add(native.second, nano_div(native.microsecond, 1000000))

    def test_iso_format(self):
        t = Time(12, 34, 56, 789123456)
        assert "12:34:56.789123456" == t.iso_format()

    def test_iso_format_with_trailing_zeroes(self):
        t = Time(12, 34, 56, 789000000)
        assert "12:34:56.789000000" == t.iso_format()

    def test_iso_format_with_leading_zeroes(self):
        t = Time(12, 34, 56, 789)
        assert "12:34:56.000000789" == t.iso_format()

    def test_from_iso_format_hour_only(self):
        expected = Time(12, 0, 0)
        actual = Time.from_iso_format("12")
        assert expected == actual

    def test_from_iso_format_hour_and_minute(self):
        expected = Time(12, 34, 0)
        actual = Time.from_iso_format("12:34")
        assert expected == actual

    def test_from_iso_format_hour_minute_second(self):
        expected = Time(12, 34, 56)
        actual = Time.from_iso_format("12:34:56")
        assert expected == actual

    def test_from_iso_format_hour_minute_second_milliseconds(self):
        expected = Time(12, 34, 56, 123000000)
        actual = Time.from_iso_format("12:34:56.123")
        assert expected == actual

    def test_from_iso_format_hour_minute_second_microseconds(self):
        expected = Time(12, 34, 56, 123456000)
        actual = Time.from_iso_format("12:34:56.123456")
        assert expected == actual

    def test_from_iso_format_hour_minute_second_nanosecond(self):
        expected = Time(12, 34, 56, 123456789)
        actual = Time.from_iso_format("12:34:56.123456789")
        assert expected == actual

    def test_from_iso_format_with_positive_tz(self):
        expected = Time(12, 34, 56, 123456789, tzinfo=FixedOffset(754))
        actual = Time.from_iso_format("12:34:56.123456789+12:34")
        assert expected == actual

    def test_from_iso_format_with_negative_tz(self):
        expected = Time(12, 34, 56, 123456789, tzinfo=FixedOffset(-754))
        actual = Time.from_iso_format("12:34:56.123456789-12:34")
        assert expected == actual

    def test_from_iso_format_with_positive_long_tz(self):
        expected = Time(12, 34, 56, 123456789, tzinfo=FixedOffset(754))
        actual = Time.from_iso_format("12:34:56.123456789+12:34:56.123456")
        assert expected == actual

    def test_from_iso_format_with_negative_long_tz(self):
        expected = Time(12, 34, 56, 123456789, tzinfo=FixedOffset(-754))
        actual = Time.from_iso_format("12:34:56.123456789-12:34:56.123456")
        assert expected == actual

    def test_from_iso_format_with_hour_only_tz(self):
        expected = Time(12, 34, 56, 123456789, tzinfo=FixedOffset(120))
        actual = Time.from_iso_format("12:34:56.123456789+02:00")
        assert expected == actual

    def test_utc_offset_fixed(self):
        expected = Time(12, 34, 56, 123456789, tzinfo=FixedOffset(-754))
        actual = -754 * 60
        assert expected.utc_offset().total_seconds() == actual

    def test_utc_offset_variable(self):
        expected = Time(12, 34, 56, 123456789, tzinfo=FixedOffset(-754))
        actual = -754 * 60
        assert expected.utc_offset().total_seconds() == actual

    def test_iso_format_with_time_zone_case_1(self):
        # python -m pytest tests/unit/time/test_time.py -s -v -k test_iso_format_with_time_zone_case_1
        expected = Time(7, 54, 2, 129790999, tzinfo=timezone_utc)
        assert expected.iso_format() == "07:54:02.129790999+00:00"
        assert expected.tzinfo == FixedOffset(0)
        actual = Time.from_iso_format("07:54:02.129790999+00:00")
        assert expected == actual

    def test_iso_format_with_time_zone_case_2(self):
        # python -m pytest tests/unit/time/test_time.py -s -v -k test_iso_format_with_time_zone_case_2
        expected = Time.from_iso_format("07:54:02.129790999+01:00")
        assert expected.tzinfo == FixedOffset(60)
        assert expected.iso_format() == "07:54:02.129790999+01:00"

    def test_to_native_case_1(self):
        # python -m pytest tests/unit/time/test_time.py -s -v -k test_to_native_case_1
        t = Time(12, 34, 56, 789123456)
        native = t.to_native()
        assert native.hour == t.hour
        assert native.minute == t.minute
        assert nano_add(native.second, nano_div(native.microsecond, 1000000)) \
               == 56.789123
        assert native.tzinfo is None
        assert native.isoformat() == "12:34:56.789123"

    def test_to_native_case_2(self):
        # python -m pytest tests/unit/time/test_time.py -s -v -k test_to_native_case_2
        t = Time(12, 34, 56, 789123456, tzinfo=timezone_utc)
        native = t.to_native()
        assert native.hour == t.hour
        assert native.minute == t.minute
        assert nano_add(native.second, nano_div(native.microsecond, 1000000)) \
               == 56.789123
        assert native.tzinfo == FixedOffset(0)
        assert native.isoformat() == "12:34:56.789123+00:00"

    def test_from_native_case_1(self):
        # python -m pytest tests/unit/time/test_time.py -s -v -k test_from_native_case_1
        native = time(12, 34, 56, 789123)
        t = Time.from_native(native)
        assert t.hour == native.hour
        assert t.minute == native.minute
        assert t.second == Decimal(native.microsecond) / 1000000 + native.second
        assert t.tzinfo is None

    def test_from_native_case_2(self):
        # python -m pytest tests/unit/time/test_time.py -s -v -k test_from_native_case_2
        native = time(12, 34, 56, 789123, FixedOffset(0))
        t = Time.from_native(native)
        assert t.hour == native.hour
        assert t.minute == native.minute
        assert t.second == Decimal(native.microsecond) / 1000000 + native.second
        assert t.tzinfo == FixedOffset(0)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# tag::result-consume-import[]
# end::result-consume-import[]

# python -m pytest tests/integration/examples/test_result_consume_example.py -s -v

def result_consume_example(driver):
    with driver.session() as session:
        session.run("MATCH (_) DETACH DELETE _").consume()

    with driver.session() as session:
        session.run("CREATE (a:Person {name: $name}) RETURN a", name="Alice").single().value()
        session.run("CREATE (a:Person {name: $name}) RETURN a", name="Bob").single().value()

    # tag::result-consume[]
    def match_person_nodes(tx):
        result = tx.run("MATCH (a:Person) RETURN a.name ORDER BY a.name")
        return [record["a.name"] for record in result]

    with driver.session() as session:
        people = session.read_transaction(match_person_nodes)
    # end::result-consume[]

    with driver.session() as session:
        session.run("MATCH (_) DETACH DELETE _").consume()

    return people


def test_example(driver):
    people = result_consume_example(driver)
    assert people == ['Alice', 'Bob']



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest


# TODO: this test will stay until a uniform behavior for `.single()` across the
#       drivers has been specified and tests are created in testkit
def test_result_single_with_no_records(session):
    result = session.run("CREATE ()")
    record = result.single()
    assert record is None


# TODO: this test will stay until a uniform behavior for `.single()` across the
#       drivers has been specified and tests are created in testkit
def test_result_single_with_one_record(session):
    result = session.run("UNWIND [1] AS n RETURN n")
    record = result.single()
    assert record["n"] == 1


# TODO: this test will stay until a uniform behavior for `.single()` across the
#       drivers has been specified and tests are created in testkit
def test_result_single_with_multiple_records(session):
    import warnings
    result = session.run("UNWIND [1, 2, 3] AS n RETURN n")
    with pytest.warns(UserWarning, match="Expected a result with a single record"):
        record = result.single()
        assert record[0] == 1


# TODO: this test will stay until a uniform behavior for `.single()` across the
#       drivers has been specified and tests are created in testkit
def test_result_single_consumes_the_result(session):
    result = session.run("UNWIND [1, 2, 3] AS n RETURN n")
    with pytest.warns(UserWarning, match="Expected a result with a single record"):
        _ = result.single()
        records = list(result)
        assert records == []
